import coeffstore
import caching
import libadjoint
import dolfin
import lusolver

# Create the adjointer, the central object that records the forward solve
# as it happens.
adjointer = libadjoint.Adjointer()

mem_checkpoints = set()
disk_checkpoints = set()

adj_variables = coeffstore.CoeffStore()

def adj_start_timestep(time=0.0):
  '''Dolfin does not supply us with information about timesteps, and so more information
  is required from the user for certain features. This function should be called at the
  start of the time loop with the initial time (defaults to 0).
  
  See also: :py:func:`dolfin_adjoint.adj_inc_timestep`
  '''

  if not dolfin.parameters["adjoint"]["stop_annotating"]:
    adjointer.time.start(time)

def adj_inc_timestep(time=None, finished=False):
  '''Dolfin does not supply us with information about timesteps, and so more information
  is required from the user for certain features. This function should be called at
  the end of the time loop with two arguments:

    - :py:data:`time` -- the time at the end of the timestep just computed
    - :py:data:`finished` -- whether this is the final timestep.

  With this information, complex functional expressions using the :py:class:`Functional` class
  can be used.

  The finished argument is necessary because the final step of a functional integration must perform
  additional calculations.

  See also: :py:func:`dolfin_adjoint.adj_start_timestep`
  '''

  if not dolfin.parameters["adjoint"]["stop_annotating"]:
    adj_variables.increment_timestep()
    if time:
      adjointer.time.next(time)

    if finished:
      adjointer.time.finish()

# A dictionary that saves the functionspaces of all checkpoint variables that have been saved to disk
checkpoint_fs = {}

function_names = set()

def adj_check_checkpoints():
  adjointer.check_checkpoints()

lu_solvers = caching.LUCache()

def adj_reset_cache():
  if dolfin.parameters["adjoint"]["debug_cache"]:
    dolfin.info_blue("Reseting solver cache")

  lu_solvers.clear()

  dolfin.parameters["adjoint"]["stop_annotating"] = False

def adj_html(*args, **kwargs):
  '''This routine dumps the current state of the adjglobals.adjointer to a HTML visualisation.
  Use it like:

    - adj_html("forward.html", "forward") # for the equations recorded on the forward run
    - adj_html("adjoint.html", "adjoint") # for the equations to be assembled on the adjoint run
  '''
  return adjointer.to_html(*args, **kwargs)

def adj_reset():
  '''Forget all annotation, and reset the entire dolfin-adjoint state.'''
  adjointer.reset()
  adj_variables.__init__()
  function_names.__init__()
  lusolver.lu_solvers = {}
  lusolver.adj_lu_solvers = {}
  adj_reset_cache()

# Map from FunctionSpace to LUSolver that has factorised the fsp mass matrix
fsp_lu = {}
